using System;

namespace ImageSharp.Maths
{
	/// <summary>
	/// Speicherstruktur und Methoden fr komplexe Zahlen.
	/// </summary>
    public struct Complex : IComparable<Complex>, IEquatable<Complex>
    {
        #region ffentliche Attribute und Eigenschaften
        /// <summary>
        /// Realteil der komplexen Zahl.
        /// </summary>
		public float Re;
        /// <summary>
        ///  Imaginrteil der komplexen Zahl.
        /// </summary>
		public float Im;
		
		/// <summary>
		/// Liefert eine komplexe Zahl mit dem Wert 0 zurck.
		/// </summary>
		public static Complex Zero
		{
			get { return new Complex(0, 0); }
		}
		
		/// <summary>
		/// Erhalte den Betrag der komplexen Zahl.
		/// </summary>
		public float Magnitude
		{
			get { return (float) System.Math.Sqrt(Re * Re + Im * Im); }
		}

		/// <summary>
		/// Erhalte die Phase der komplexen Zahl.
		/// </summary>
		public float Phase
		{
			get { return (float) System.Math.Atan(Im / Re); }
		}
		
		/// <summary>
		/// Erhalte den quadrierten Betrag der komplexen Zahl.
		/// </summary>
		public float SquaredMagnitude
		{
			get { return (Re * Re + Im * Im); }
        }
        #endregion

        #region ffentliche Methoden
        public Complex(float re, float im)
		{
			this.Re = re;
			this.Im = im;
		}
		public Complex(Complex c)
		{
			this.Re = c.Re;
			this.Im = c.Im;
		}
		
		/// <summary>
		/// Wandle die komplexe Zahl in einen String um.
		/// </summary>
		/// <returns>String der komplexen Zahl</returns>
		public override string ToString()
		{
			return String.Format("{0}{1}{2}i", Re, ((Im < 0) ? '-' : '+'), Math.Abs(Im));
        }

        #region Arithmetische Operationen mit komplexen Zahlen
        /// <summary>
		/// Addition zweier komplexer Zahlen.
		/// </summary>
		/// <param name="a">Erste komplexe Zahl</param>
		/// <param name="b">Zweite komplexe Zahl</param>
		/// <returns>Summe von a und b</returns>
		public static Complex operator+(Complex a, Complex b)
		{
			return new Complex(a.Re + b.Re, a.Im + b.Im);
		}

		/// <summary>
        /// Subtrahiere zwei komplexe Zahlen voneinander.
		/// </summary>
		/// <param name="a">Erste komplexe Zahl</param>
		/// <param name="b">Zweite komplexe Zahl</param>
		/// <returns>Differenz von a und b</returns>
		public static Complex operator-(Complex a, Complex b)
		{
			return new Complex(a.Re - b.Re, a.Im - b.Im);
		}
		
		/// <summary>
		/// Multipliziere zwei komplexe Zahlen miteinander.
		/// </summary>
        /// <param name="a">Erste komplexe Zahl</param>
        /// <param name="b">Zweite komplexe Zahl</param>
        /// <returns>a * b</returns>
		public static Complex operator*(Complex a, Complex b)
		{
			return new Complex(
				a.Re * b.Re - a.Im * b.Im,
				a.Re * b.Im + a.Im * b.Re);
		}

		/// <summary>
		/// Division zweier komplexer Zahlen.
		/// </summary>
        /// <param name="a">Erste komplexe Zahl</param>
        /// <param name="b">Zweite komplexe Zahl</param>
        /// <returns>Quotient von a und b</returns>
		public static Complex operator/(Complex a, Complex b)
		{
			float divider = b.Re * b.Re + b.Im * b.Im;
			
			if (divider == 0)
				throw new DivideByZeroException();
		
			return new Complex(
				(a.Re * b.Re + a.Im * b.Im) / divider,
				(a.Im * b.Re - a.Re * b.Im) / divider);
		}

        /// <summary>
        /// Addition einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">Integer-Wert</param>
        /// <returns>a + b</returns>
        public static Complex operator +(Complex a, int b)
        {
            return new Complex(a.Re + b, a.Im + b);
        }

        /// <summary>
        /// Addition einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Integer-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a + b</returns>
        public static Complex operator +(int a, Complex b)
        {
            return b + a;
        }

        /// <summary>
        /// Subtraktion einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">Integer-Wert</param>
        /// <returns>a - b</returns>
        public static Complex operator -(Complex a, int b)
        {
            return new Complex(a.Re - b, a.Im - b);
        }

        /// <summary>
        /// Subtraktion einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Integer-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a - b</returns>
        public static Complex operator -(int a, Complex b)
        {
            return new Complex(a - b.Re, a - b.Im);
        }

        /// <summary>
        /// Multiplikation einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">Integer-Wert</param>
        /// <returns>a * b</returns>
        public static Complex operator *(Complex a, int b)
        {
            return new Complex(a.Re * b, a.Im * b);
        }

        /// <summary>
        /// Multiplikation einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Integer-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a * b</returns>
        public static Complex operator *(int a, Complex b)
        {
            return b * a;
        }

        /// <summary>
        /// Division einer komplexen Zahl und eines Integer-Wertes.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">Integer-Wert</param>
        /// <returns>a / b</returns>
        public static Complex operator /(Complex a, int b)
        {
            if (b == 0)
                throw new DivideByZeroException();

            return new Complex(a.Re / b, a.Im / b);
        }

        /// <summary>
        /// Division einer komplexen Zahl mit einem Integer-Wert.
        /// </summary>
        /// <param name="a">Integer-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a / b</returns>
        public static Complex operator /(int a, Complex b)
        {
            if (b.Re == 0 || b.Im == 0)
                throw new DivideByZeroException();

            return new Complex(a / b.Re, a / b.Im);
        }

        /// <summary>
        /// Addition einer komplexen Zahl mit einem float-Wert.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">float-Wert</param>
        /// <returns>a + b</returns>
        public static Complex operator +(Complex a, float b)
        {
            return new Complex(a.Re + b, a.Im + b);
        }

        /// <summary>
        /// Addition einer komplexen Zahl mit einem float-Wert.
        /// </summary>
        /// <param name="a">float-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a + b</returns>
        public static Complex operator +(float a, Complex b)
        {
            return b + a;
        }

        /// <summary>
        /// Subtraktion einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">float-Wert</param>
        /// <returns>a - b</returns>
        public static Complex operator -(Complex a, float b)
        {
            return new Complex(a.Re - b, a.Im - b);
        }

        /// <summary>
        /// Subtraktion einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">float-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a - b</returns>
        public static Complex operator -(float a, Complex b)
        {
            return new Complex(a - b.Re, a - b.Im);
        }

        /// <summary>
        /// Multiplikation einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">float-Wert</param>
        /// <returns>a * b</returns>
        public static Complex operator *(Complex a, float b)
        {
            return new Complex(a.Re * b, a.Im * b);
        }

        /// <summary>
        /// Multiplikation einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">float-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns>a * b</returns>
        public static Complex operator *(float a, Complex b)
        {
            return b * a;
        }

        /// <summary>
        /// Division einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">Komplexe Zahl</param>
        /// <param name="b">float-Wert</param>
        /// <returns></returns>
        public static Complex operator /(Complex a, float b)
        {
            if (b == 0)
                throw new DivideByZeroException();

            return new Complex(a.Re / b, a.Im / b);
        }

        /// <summary>
        /// Division einer komplexen Zahl und eines float-Wertes.
        /// </summary>
        /// <param name="a">float-Wert</param>
        /// <param name="b">Komplexe Zahl</param>
        /// <returns></returns>
        public static Complex operator /(float a, Complex b)
        {
            if (b.Re == 0 || b.Im == 0)
                throw new DivideByZeroException();

            return new Complex(a / b.Re, a / b.Im);
        }
        #endregion
        #endregion

        #region interface implementations
        public int CompareTo(Complex other)
        {
            int c = this.Re.CompareTo(other.Re);
            if (c == 0)
            {
                return this.Im.CompareTo(other.Im);
            }
            else
            {
                return c;
            }
        }

        public bool Equals(Complex other)
        {
            return (this.Re.Equals(other.Re) && this.Im.Equals(other.Im));
        }
        #endregion
    }
}