using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;

namespace ImageSharp.Images
{
    /// <summary>
    /// Klasse fr ein Farbbild mit zugehrigen Operationen.
    /// </summary>
    class ImageColor : IImage
    {
        #region Private Attribute
        private int height = 0, width = 0;  // Hhe und Breite des Bildes
        private RGBData[,] data = null;   // Pixelwerte
        #endregion

        /// <summary>
        /// Gibt die RGB-Daten der einzelnen Pixel zurck.
        /// </summary>
        public RGBData[,] Data
        {
            get
            {
                return this.data;
            }
        }

        /// <summary>
        /// Gibt die Bildhhe zurck.
        /// </summary>
        public int Height
        {
            get
            {
                return this.height;
            }
        }

        /// <summary>
        /// Gibt die Bildbreite zurck.
        /// </summary>
        public int Width
        {
            get
            {
                return this.width;
            }
        }

        /// <summary>
        /// Erzeugt ein neues Farbbild als tiefe Kopie des bergebenen Farbbildes.
        /// </summary>
        /// <param name="img">Zu klonendes Farbbild</param>
        public ImageColor(ImageColor img)
        {
            this.height = img.Height;
            this.width = img.Width;
            this.data = (RGBData[,])img.Data.Clone();
        }

        /// <summary>
        /// Erzeugt ein Farbbild mit dem bergebenen Grauwertbild.
        /// </summary>
        /// <param name="img">Zu bernehmendes Grauwertbild</param>
        public ImageColor(ImageGray img)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = img.Height;
            this.width = img.Width;
            this.data = new RGBData[this.height, this.width];

            for (int y = 0; y < this.height; y++)
            {
                for (int x = 0; x < this.width; x++)
                {
                    this.data[y, x].R = img.Data[y, x];
                    this.data[y, x].G = img.Data[y, x];
                    this.data[y, x].B = img.Data[y, x];
                    this.data[y, x].A = (byte)0;
                }
            }
        }

        /// <summary>
        /// Erzeugt ein Farbbild aus dem bergebenen .NET Bitmap-Objekt.
        /// </summary>
        /// <param name="img">Bild, dessen Farbbild erzeugt werden soll</param>
        public unsafe ImageColor(Bitmap img)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = img.Height;
            this.width = img.Width;
            this.data = new RGBData[this.height, this.width];

            // Pixeldaten kopieren
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = img.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = img.LockBits(
                bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            for (int y = 0; y < bounds.Height; y++)
            {
                for (int x = 0; x < bounds.Width; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    // RGB-Werte bernehmen
                    data[y, x].R = pPixel->R;
                    data[y, x].G = pPixel->G;
                    data[y, x].B = pPixel->B;
                    data[y, x].A = pPixel->A;
                }
            }

            img.UnlockBits(bitData);
        }

        /// <summary>
        /// Erzeugt ein Farbbild mit den bergebenen RGB-Daten. Dieses wird nicht tief kopiert,
        /// sondern nur die Referenz darauf bernommen.
        /// </summary>
        /// <param name="data">Zu bernehmendes Daten-Array</param>
        public ImageColor(RGBData[,] data)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = data.GetLength(0);
            this.width = data.GetLength(1);
            this.data = data;
        }

        /// <summary>
        /// Erzeugt ein "leeres" Farbbild in der bergebenen Gre.
        /// </summary>
        /// <param name="width">Bildbreite</param>
        /// <param name="height">Bildhhe</param>
        public ImageColor(int width, int height)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = height;
            this.width = width;
            this.data = new RGBData[this.height, this.width];
        }

        /// <summary>
        /// Erzeugt ein .NET Bitmap Objekt aus den hier enthaltenen RGB-Daten.
        /// </summary>
        /// <returns>Bitmap-Objekt</returns>
        public unsafe Bitmap ToBitmap()
        {
            // Erzeuge das Bitmap
            Bitmap dstImg = new Bitmap(this.width, this.height);

            // Kopiere die Daten
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = dstImg.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = dstImg.LockBits(
                bounds, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            for (int y = 0; y < bounds.Height; y++)
            {
                for (int x = 0; x < bounds.Width; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    pPixel->R = this.data[y, x].R;
                    pPixel->G = this.data[y, x].G;
                    pPixel->B = this.data[y, x].B;
                    pPixel->A = this.data[y, x].A;
                }
            }

            dstImg.UnlockBits(bitData);
            return dstImg;
        }

        public IImage Clone()
        {
            return new ImageColor(this);
        }

        public void Apply(IColorImgApplicable op)
        {
            op.Apply(this);
        }

        #region interface implementations
        public IPixelData GetValueAt(int x, int y)
        {
            return this.data[y, x];
        }

        public void SetValueAt(int x, int y, IPixelData value)
        {
            this.data[y, x] = (RGBData)value;
        }

        public void CopyFrom(ImageColor img)
        {
            this.data = new RGBData[img.height, img.width];
            for (int y = 0; y < img.height; y++)
                for (int x = 0; x < img.width; x++)
                    this.data[y, x] = img.data[y, x];
            this.height = img.height;
            this.width = img.width;
        }

        public void CopyFrom(RGBData[,] data)
        {
            int height = data.GetLength(0);
            int width = data.GetLength(1);
            this.data = new RGBData[height, width];
            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                    this.data[y, x] = data[y, x];
            this.height = height;
            this.width = width;
        }

        public void CopyFrom(IImage img)
        {
            ImageColor nimg = null;
            if (img is ImageColor)
                nimg = new ImageColor((ImageColor)img);
            else if (img is ImageGray)
                nimg = new ImageColor((ImageGray)img);
            else if (img is ImageComplex)
                nimg = new ImageColor(new ImageGray((ImageComplex)img));

            this.data = (RGBData[,])nimg.data.Clone();
            this.height = nimg.height;
            this.width = nimg.width;
        }

        public void CreateFrom(int width, int height)
        {
            this.width = width;
            this.height = height;
            this.data = new RGBData[height, width];
        }
        #endregion
    }
}
