using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using ImageSharp.Images;
using ImageSharp.Maths;

namespace ImageSharp.Filtering.Linear
{
    /// <summary>
    /// Stellt Methoden zur Anwendung linearer Filter ber Filtermatrizen zur Verfgung.
    /// </summary>
    class LinearFilter : IFilter
    {
        #region Private/Protected Attribute
        // Anzuwendende Filtermatrix.
        protected double[,] filterMatrix;
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Gibt die anzuwendende Filtermatrix zurck.
        /// </summary>
        public double[,] FilterMatrix
        {
            get
            {
                return filterMatrix;
            }
        }
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Erzeugt einen linearen Filter mit der bergebenen Filtermatrix.
        /// </summary>
        /// <param name="matrix">Anzuwendende Filtermatrix</param>
        public LinearFilter(double[,] matrix)
        {
            this.filterMatrix = matrix;
        }

        /// <summary>
        /// Erzeugt einen linearen Filter mit der bergebenen Filtermatrix.
        /// </summary>
        /// <param name="matrix">Anzuwendende Filtermatrix</param>
        public LinearFilter(int[,] matrix)
        {
            // Wandle int-Matrix in double um.
            this.filterMatrix = new double[matrix.GetLength(0), matrix.GetLength(1)];
            for (int i = 0; i < matrix.GetLength(0); i++)
            {
                for (int j = 0; j < matrix.GetLength(1); j++)
                {
                    this.filterMatrix[i, j] = (double)matrix[i, j];
                }
            }
        }

        /// <summary>
        /// Wendet die Filtermatrix auf das bergebene Bild im eingegrenzten Bildraum an.
        /// </summary>
        /// <param name="img">Zu bearbeitendes Bild</param>
        public virtual void Apply(ImageGray img)
        {
            byte[,] dataCopy = (byte[,])img.Data.Clone();

            // Additions- und Divisionskomponenten berechen;
            // diese werden zur Skalierung des Ergebnisses auf den Grauwertbereich [0;255] bentigt.
            double div = 0.0, add = 0.0;
            for (int i = 0; i < filterMatrix.GetLength(0); i++)
            {
                for (int j = 0; j < filterMatrix.GetLength(1); j++)
                {
                    // Division muss durch die Summe der Elemente der Filtermatrix stattfinden.
                    div += Math.Abs(filterMatrix[i, j]);
                    if (filterMatrix[i, j] < 0)
                    {
                        // Addition findet mit der Summer der negativen Filtermatrix-Elemente statt.
                        add -= filterMatrix[i, j];
                    }
                }
            }

            // Wende fr jedes Pixel an
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    // Berechne Summe der Filtermatrix * Pixelwerte mit Bezugspixel in der Mitte.
                    double pxSum = 0.0;
                    for (int i = 0; i < filterMatrix.GetLength(0); i++)
                    {
                        for (int j = 0; j < filterMatrix.GetLength(1); j++)
                        {
                            int x1 = x + j - filterMatrix.GetLength(1) / 2;
                            int y1 = y + i - filterMatrix.GetLength(0) / 2;
                            if (x1 >= 0 && x1 < img.Width && y1 >= 0 && y1 < img.Height)
                            {
                                pxSum += (double)dataCopy[y1, x1] * filterMatrix[i, j];
                            }
                        }
                    }
                    // Fhre Skalierung auf Grauwertbereich durch.
                    if (div != 0.0)
                    {
                        img.Data[y, x] = (byte)((pxSum + 255.0 * add) / div);
                    }
                }
            }
        }
        #endregion
    }

    /// <summary>
    /// Stellt einen adaptiven Gauss-Filter dar, bei dem die Standardabweichung eingestellt werden kann.
    /// </summary>
    class AdaptiveGaussFilter : LinearFilter
    {
        #region ffentliche Methoden
        /// <summary>
        /// Erzeugt einen Gauss-Filter mit einer Standardabweichung von 1.0.
        /// </summary>
        public AdaptiveGaussFilter()
            : base(MathCalc.GaussMask(1.0))
        { }

        /// <summary>
        /// Erzeugt einen Gauss-Filter mit der bergebenen Standardabweichung.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        public AdaptiveGaussFilter(double sigma)
            : base(MathCalc.GaussMask(sigma))
        { }
        #endregion
    }

    /// <summary>
    /// Diese Klasse stellt statische get-Properties fr diverse lineare Filter zur Verfgung.
    /// </summary>
    sealed class LinearFilters
    {
        #region ffentliche Eigenschaften
        /// <summary>
        /// Gibt ein Objekt des adaptiven Gauss-Filters mit einer Standardabweichung von 1.0 zurck.
        /// </summary>
        public static AdaptiveGaussFilter AdaptiveGauss
        {
            get
            {
                return new AdaptiveGaussFilter(1.0);
            }
        }
        
        /// <summary>
        /// Gibt ein Objekt des Gauss-Tiefpassfilters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter Gaussian
        {
            get
            {
                return new LinearFilter(new double[,] {
                    { 1, 2, 1 }, 
                    { 2, 4, 2 }, 
                    { 1, 2, 1 } });
            }
        }

        /// <summary>
        /// Gibt ein Objekt des homogenen Tiefpassfilters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter Lowpass
        {
            get
            {
                return new LinearFilter(new double[,] {
                    { 1, 1, 1 }, 
                    { 1, 1, 1 }, 
                    { 1, 1, 1 } });
            }
        }

        /// <summary>
        /// Gibt ein Objekt eines Hochpassfilters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter Highpass
        {
            get
            {
                return new LinearFilter(new double[,] {
                    { -1, -1, -1 }, 
                    { -1,  8, -1 }, 
                    { -1, -1, -1 } });
            }
        }

        /// <summary>
        /// Gibt ein Objekt eines Highboost-Filters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter Highboost
        {
            get
            {
                double A = 2.0;
                return new LinearFilter(new double[,] {
                    { -1,    -1, -1 }, 
                    { -1, 9*A-1, -1 }, 
                    { -1,    -1, -1 } });
            }
        }

        /// <summary>
        /// Gibt ein Objekt des Laplace-Filters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter Laplacian
        {
            get
            {
                return new LinearFilter(new double[,] {
                    {  0, -1,  0 }, 
                    { -1,  4, -1 }, 
                    {  0, -1,  0 } });
            }
        }

        /// <summary>
        /// Gibt ein Objekt des Laplace-Schrfungsfilters mit einer 3x3-Filtermatrix zurck.
        /// </summary>
        public static LinearFilter LaplacianSharpen
        {
            get
            {
                return new LinearFilter(new double[,] {
                    {  0, -1,  0 }, 
                    { -1,  5, -1 }, 
                    {  0, -1,  0 } });
            }
        }
        #endregion
    }
}
