using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ImageSharp
{
    /// <summary>
    /// Speicherstruktur fr eine Gerade im Hough-Array. Speichert neben der eigentlichen Linie auch
    /// deren Betrag im Akku-Array und ihre Lnge sowie die Anzahl der Pixel, die auf ihr liegen.
    /// </summary>
    class HoughLine
    {
        #region ffentliche Attribute und Eigenschaften
        /// <summary>
        /// Die Linie, welche dem gespeicherten Akku-Element entspricht.
        /// </summary>
        public Line Line = null;

        /// <summary>
        /// Der Betrag der Geraden im Hough-Akku.
        /// </summary>
        public double Accuval = 0.0;

        /// <summary>
        /// Die Anzahl der Pixel, die auf der Geraden liegen.
        /// </summary>
        public double PxCount = 0.0;

        /// <summary>
        /// Die Lnge der Geraden.
        /// </summary>
        public double LineLength = 0.0;

        /// <summary>
        /// Der normierte Akku-Betrag (multipliziert mit der Anzahl der Pixel dividiert durch die
        /// Lnge der Geraden).
        /// </summary>
        public double NormedAccuval
        {
            get
            {
                return (LineLength == 0) ? 0 : Accuval * PxCount / LineLength;
            }
        }
        #endregion

        #region Konstruktoren
        public HoughLine() { }
        public HoughLine(Line line, double accuval, double pxCount, double lineLength)
        {
            this.Line = line;
            this.Accuval = accuval;
            this.PxCount = pxCount;
            this.LineLength = lineLength;
        }
        #endregion
    }

    /// <summary>
    /// Speicherstruktur einer Geraden.
    /// </summary>
    class Line
    {
        #region ffentliche Attribute und Eigenschaften
        /// <summary>
        /// Startpunkt der Linie. Erst verfgbar, nachdem er mit GetStartPoint() gesetzt wurde.
        /// </summary>
        public Point StartPoint = new Point();

        /// <summary>
        /// Enpunkt der Linie. Erst verfgbar, nachdem er mit GetEndPoint() gesetzt wurde.
        /// </summary>
        public Point EndPoint = new Point();

        /// <summary>
        /// Winkel der Geraden in der Hesse'schen Normalform.
        /// </summary>
        public double Angle = -1.0;

        /// <summary>
        /// Radius der Geraden in der Hesse'schen Normalform.
        /// </summary>
        public double Radius = 0.0;

        /// <summary>
        /// Berechnet die Lnge der Geraden und gibt diese zurck. Dazu sind zunchst der Geradenstart-
        /// und -endpunkt mittels GetStartPoint() und GetEndPoint() zu setzen.
        /// </summary>
        public double Length
        {
            get
            {
                if (!StartPoint.IsEmpty && !EndPoint.IsEmpty)
                {
                    int dx = EndPoint.X - StartPoint.X;
                    int dy = EndPoint.Y - StartPoint.Y;
                    return Math.Sqrt(dx * dx + dy * dy);
                }
                else
                {
                    return 0.0;
                }
            }
        }
        #endregion

        #region ffentliche Methoden
        public Line() { }

        public Line(Point start, Point end)
        {
            this.StartPoint = start;
            this.EndPoint = end;
        }

        public Line(double angle, double radius)
        {
            this.Angle = angle;
            this.Radius = radius;
        }

        public Line(double angle, double radius, Point start, Point end)
        {
            this.StartPoint = start;
            this.EndPoint = end;
            this.Angle = angle;
            this.Radius = radius;
        }

        /// <summary>
        /// Startpunkt der Linie ermitteln und im Attribut StartPoint eintragen.
        /// Dazu wird der Schnittpunkt der Geraden mit dem ersten Element im bergebenen
        /// Bildraum berechnet.
        /// </summary>
        /// <param name="limits">Eingrenzender Bildraum</param>
        /// <returns>Startpunkt der Geraden als Schnittpunkt mit der oberen Bildbereichsgrenze</returns>
        /*public Point GetStartPoint(ImageLimiter[] limits)
        {
            // Schnittpunkt der Geraden mit der obersten Teilbereichsgrenze ermitteln
            int ymin = limits[0].Y;
            StartPoint = new Point(MathCalc.GetInterceptCoord(Angle, Radius, ymin), ymin);

            // Schnittpunkt zurckgeben
            return StartPoint;
        }*/

        /// <summary>
        /// Endpunkt der Linie ermitteln und im Attribut EndPoint eintragen.
        /// Dazu wird der Schnittpunkt der Geraden mit dem letzten Element im bergebenen
        /// Bildraum berechnet.
        /// </summary>
        /// <param name="limits">Eingrenzender Bildraum</param>
        /// <returns>Endpunkt der Geraden als Schnittpunkt mit der unteren Bildbereichsgrenze</returns>
        /*public Point GetEndPoint(ImageLimiter[] limits)
        {
            // Schnittpunkt der Geraden mit der untersten Teilbereichsgrenze ermitteln
            int ymax = limits[limits.Length - 1].Y;
            EndPoint = new Point(MathCalc.GetInterceptCoord(Angle, Radius, ymax), ymax);

            // Schnittpunkt zurckgeben
            return EndPoint;
        }*/

        /// <summary>
        /// Statische Methode, die ein neues Line-Objekt anhand der bergebenen Punkte erstellt,
        /// die auf der Geraden liegen sollen.
        /// </summary>
        /// <param name="start"></param>
        /// <param name="end"></param>
        /// <returns></returns>
        public static Line FromPoints(Point start, Point end)
        {
            // Winkel berechnen und in ein Intervall von [0,PI] berfhren
            double angle = Math.PI/2.0 + Math.Atan2(end.Y - start.Y, end.X - start.X);
            if (angle < 0)
                angle = Math.PI + angle;
            if (angle > Math.PI)
                angle = angle - Math.PI;

            // entsprechenden Radius berechnen
            double radius = start.X * Math.Cos(angle) + start.Y * Math.Sin(angle);

            // neues Line-Objekt zurckgeben
            return new Line(angle, radius, start, end);
        }
        #endregion
    }
}
