using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;

namespace ImageSharp.Filtering.Spectral
{
    abstract class SpectralFilter : IFilter, IComplexImgApplicable
    {
        public void Apply(ImageComplex img)
        {
            if (img.Fmode)
            {
                applyFilter(img);
            }
            else
            {
                img.MakeForwardFourierTransform();
                applyFilter(img);
                img.MakeBackwardFourierTransform();
            }
        }

        public void Apply(ImageGray img)
        {
            ImageComplex imgComplex = new ImageComplex(img);
            imgComplex.MakeForwardFourierTransform();
            applyFilter(imgComplex);
            imgComplex.MakeBackwardFourierTransform();
            ImageGray imgGray = new ImageGray(imgComplex);
            img.Data = imgGray.Data;
        }

        protected abstract void applyFilter(ImageComplex img);
    }

    /// <summary>
    /// Anwendung der Laplace-Funktion, um eine Hochpassfilterung zu erreichen.
    /// </summary>
    class LaplacianSpectral : SpectralFilter
    {
        protected override void applyFilter(ImageComplex img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                int y1 = y - img.Height / 2;
                for (int x = 0; x < img.Width; x++)
                {
                    // Zentrumskoordinaten berechnen
                    int x1 = x - img.Width / 2;
                    // Laplace-Wert berechnen
                    double t1 = 4f * Math.PI * Math.PI;
                    double t2 = (x1 * x1 + y1 * y1);
                    float d2h = (float)(-t1 * t2);
                    if (Math.Abs(d2h) > 1)
                        d2h = 1;
                    // Faltung im Ortsfrequenzraum entspricht einer einfachen Multiplikation
                    img.Data[y, x] *= d2h;
                }
            }
        }
    }

    /// <summary>
    /// LoG-Filterung anwenden, um ein bandpassgefiltertes Bild nach der Mexikanerhutfunktion (LoG) zu erhalten.
    /// </summary>
    class LaplacianOfGaussianSpectral : SpectralFilter
    {
        private double sigma = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        public LaplacianOfGaussianSpectral(double sigma)
        {
            this.sigma = sigma;
        }

        protected override void applyFilter(ImageComplex img)
        {
            double sq = this.sigma * this.sigma;
            for (int y = 0; y < img.Height; y++)
            {
                int y1 = y - img.Height / 2;
                for (int x = 0; x < img.Width; x++)
                {
                    // Zentrumskoordinaten erzeugen
                    int x1 = x - img.Width / 2;
                    // Wert der LoG-Funktion errechnen (muss wohlmglich angepasst werden)
                    double rq = (x1 * x1 + y1 * y1) / (img.Width * img.Height / 1000f);
                    double t1 = 1 - rq / sq;
                    double t2 = Math.Exp(-rq / (2 * sq));
                    float d2h = (float)(-t1 * t2);
                    // Faltung durch Multiplikation
                    img.Data[y, x] *= d2h;
                }
            }
        }
    }

    /// <summary>
    /// Ideale Bandpass-Filterung durchfhren.
    /// </summary>
    class BandpassSpectral : SpectralFilter
    {
        private int min = 0;
        private int max = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="min">Minimal durchzulassende Frequenz</param>
        /// <param name="max">Maximal durchzulassende Frequenz</param>
        public BandpassSpectral(int min, int max)
        {
            this.min = min;
            this.max = max;
        }

        protected override void applyFilter(ImageComplex img)
        {
            for (int i = 0; i < img.Height; i++)
            {
                int y = i - img.Height / 2;

                for (int j = 0; j < img.Width; j++)
                {
                    // Zentrumskoordinaten berechnen
                    int x = j - img.Width / 2;
                    // Abstand vom Mittelpunkt berechnen
                    int d = (int)Math.Sqrt(x * x + y * y);

                    // wenn auerhalb des zulssigen Bereichs, dann filtern
                    if ((d > this.max) || (d < this.min))
                    {
                        img.Data[i, j].Re = 0;
                        img.Data[i, j].Im = 0;
                    }
                }
            }
        }
    }

    /// <summary>
    /// Ideale Lowpass-Filterung durchfhren.
    /// </summary>
    class LowpassSpectral : SpectralFilter
    {
        private int cutoff = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="cutoff">Frequenz, ab der zu filtern ist</param>
        public LowpassSpectral(int cutoff)
        {
            this.cutoff = cutoff;
        }

        protected override void applyFilter(ImageComplex img)
        {
            for (int i = 0; i < img.Height; i++)
            {
                int y = i - img.Height / 2;

                for (int j = 0; j < img.Width; j++)
                {
                    // Zentrumskoordinaten berechnen
                    int x = j - img.Width / 2;
                    // Abstand vom Mittelpunkt berechnen
                    int d = (int)Math.Sqrt(x * x + y * y);

                    // wenn auerhalb des zulssigen Bereichs, dann filtern
                    if (d > this.cutoff)
                    {
                        img.Data[i, j].Re = 0;
                        img.Data[i, j].Im = 0;
                    }
                }
            }
        }
    }

    /// <summary>
    /// Ideale Highpass-Filterung durchfhren.
    /// </summary>
    class HighpassSpectral : SpectralFilter
    {
        private int cutoff = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="cutoff">Frequenz, bis zu der zu filtern ist</param>
        public HighpassSpectral(int cutoff)
        {
            this.cutoff = cutoff;
        }

        protected override void applyFilter(ImageComplex img)
        {
            for (int i = 0; i < img.Height; i++)
            {
                int y = i - img.Height / 2;

                for (int j = 0; j < img.Width; j++)
                {
                    // Zentrumskoordinaten berechnen
                    int x = j - img.Width / 2;
                    // Abstand vom Mittelpunkt berechnen
                    int d = (int)Math.Sqrt(x * x + y * y);

                    // wenn auerhalb des zulssigen Bereichs, dann filtern
                    if (d < this.cutoff)
                    {
                        img.Data[i, j].Re = 0;
                        img.Data[i, j].Im = 0;
                    }
                }
            }
        }
    }

    /// <summary>
    /// Fhrt eine geglttete Bandpass-Filterung durch.
    /// </summary>
    class SmoothBandpassFilter : SpectralFilter
    {
        private int min = 0;
        private int max = 0;
        private int omega1 = 0;
        private int omega2 = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="min">Minimal zulssige Frequenz</param>
        /// <param name="omega1">Glttungsbereich der minimalen Frequenz</param>
        /// <param name="max">Maximal zulssige Frequenz</param>
        /// <param name="omega2">Glttungsbereich der maximalen Frequenz</param>
        public SmoothBandpassFilter(int min, int omega1, int max, int omega2)
        {
            this.min = min;
            this.max = max;
            this.omega1 = omega1;
            this.omega2 = omega2;
        }

        protected override void applyFilter(ImageComplex img)
        {
            // Ideale Hoch- und Tiefpassfilterung durchfhren
            (new SmoothHighpassFilter(min, omega1)).ApplyToFourierImg(img);
            (new SmoothLowpassFilter(max, omega2)).ApplyToFourierImg(img);
        }
    }

    /// <summary>
    /// Gegltteten Tiefpassfilter anwenden.
    /// </summary>
    class SmoothLowpassFilter : SpectralFilter
    {
        private int cutoff = 0;
        private int omega = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="cutoff">Frequenz, bis zu der geglttet zu filtern ist</param>
        /// <param name="omega">Glttungsbereich um den cutoff-Wert herum</param>
        public SmoothLowpassFilter(int cutoff, int omega)
        {
            this.cutoff = cutoff;
            this.omega = omega;
        }

        public void ApplyToFourierImg(ImageComplex img)
        {
            if (img.Fmode == true)
                applyFilter(img);
        }

        protected override void applyFilter(ImageComplex img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                int y1 = y - img.Height / 2;
                for (int x = 0; x < img.Width; x++)
                {
                    int x1 = x - img.Width / 2;
                    int f = (int)Math.Sqrt(x1 * x1 + y1 * y1);
                    // Fallunterscheidung: 
                    // 1.) f < cutoff-omega: lasse alles durch
                    // 2.) f > cutoff+omega: filtere die Frequenzen
                    // 3.) dazwischen: geglttete Filterung durch Anwendung einer Sinus-Funktion
                    if (f < cutoff - omega)
                    {
                        img.Data[y, x].Re = img.Data[y, x].Re * 1;
                        img.Data[y, x].Im = img.Data[y, x].Im * 1;
                    }
                    else if (f > this.cutoff + this.omega)
                    {
                        img.Data[y, x].Re = 0;
                        img.Data[y, x].Im = 0;
                    }
                    else
                    {
                        img.Data[y, x].Re = (float)(img.Data[y, x].Re * 0.5 * (1 - Math.Sin(Math.PI * (f - this.cutoff) / (2f * this.omega))));
                        img.Data[y, x].Im = (float)(img.Data[y, x].Im * 0.5 * (1 - Math.Sin(Math.PI * (f - this.cutoff) / (2f * this.omega))));
                    }
                }
            }
        }
    }

    /// <summary>
    /// Gegltteten Hochpassfilter anwenden.
    /// </summary>
    class SmoothHighpassFilter : SpectralFilter
    {
        private int cutoff = 0;
        private int omega = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="cutoff">Frequenz, bis zu der geglttet zu filtern ist</param>
        /// <param name="omega">Glttungsbereich um den cutoff-Wert herum</param>
        public SmoothHighpassFilter(int cutoff, int omega)
        {
            this.cutoff = cutoff;
            this.omega = omega;
        }

        public void ApplyToFourierImg(ImageComplex img)
        {
            if (img.Fmode == true)
                applyFilter(img);
        }

        protected override void applyFilter(ImageComplex img)
        {
            for (int y = 0; y < img.Height; y++)
            {
                int y1 = y - img.Height / 2;
                for (int x = 0; x < img.Width; x++)
                {
                    int x1 = x - img.Width / 2;
                    // Abstand vom Bildzentrum
                    int f = (int)Math.Sqrt(x1 * x1 + y1 * y1);
                    // Fallunterscheidung: 
                    // 1.) f < cutoff-omega: filtere die Frequenzen
                    // 2.) f > cutoff+omega: lasse alles durch
                    // 3.) dazwischen: geglttete Filterung durch Anwendung einer Sinus-Funktion
                    if (f < this.cutoff - this.omega)
                    {
                        img.Data[y, x].Re = 0;
                        img.Data[y, x].Im = 0;
                    }
                    else if (f > this.cutoff + this.omega)
                    {
                        img.Data[y, x].Re = img.Data[y, x].Re * 1;
                        img.Data[y, x].Im = img.Data[y, x].Im * 1;
                    }
                    else
                    {
                        img.Data[y, x].Re = (float)(img.Data[y, x].Re * 0.5 * (1 + Math.Sin(Math.PI * (f - this.cutoff) / (2f * this.omega))));
                        img.Data[y, x].Im = (float)(img.Data[y, x].Im * 0.5 * (1 + Math.Sin(Math.PI * (f - this.cutoff) / (2f * this.omega))));
                    }
                }
            }
        }
    }
}
